/*****************************************************************************
 * Transfer.h: Main ShrinkTo5 Engine Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#ifdef __MPE2TRANSFER_DLL__
#define DLLEXPORT __declspec(dllexport) 
#else
#define DLLEXPORT __declspec(dllimport) 
#endif

/***** TransferAlert Object Section *****/

#define ALERT_PREVIEW_AVAIL		0x1		// alert is ALERT_PREVIEW_AVAIL, param is a HBITMAP. a new preview picture is available.
										// the HBITMAP present is only valid if a valid device context compatible to a display device context was provided.
										// if callback returns >= 1 the HBITMAP must be destroyed by the application that implements callback with DeleteObject
										// if callback returns <= 0 the HBITMAP is destroyed by the alerter
#define ALERT_FILENAME_CHANGED  0x2		// alert ALERT_FILENAME_CHANGED, param is LPCSTR to the new filename.
										// return value will be ignored

// this is the base class for alerts issued by the transfer object
// the application using this alert has to derive a new class implementing the callback routine
class DLLEXPORT TransferAlert
{
public:
	TransferAlert(void);
	~TransferAlert(void);

	// *** for alert receivers
	// for C++ Languages: defines a CallBack to be overriden
	virtual int CallBack(int alert, int param, void *pTransfer); 	// defines the callback routine to be called when an alert is available
																	// this function must be overwritten by a derived class
																	// the last parameter can be reinterpreted (explicit cast) to (Transfer *)
	// for non C++ Languages: sets a pointer to a callback function
	// when calling the CallBack the pTransfer parameter points to Transfer Object and pUser to the user supplied data to this function
	virtual void SetCallBack(int (_stdcall *pCallBack)(int alert, int param, void *pTransfer, void *pUser), void *pUser);

	virtual void AddAlert(int alert);			// activates one or more alerts
	virtual void SubAlert(int alert);			// deactivates one or more alerts
	virtual bool GetAlert(int alert);			// checks whether all given alerts are set (returns true if all are set, else false)
	virtual int  GetAllAlert();					// retrieves the alert flags
	virtual void SubAllAlert();					// deactivates all alerts

	virtual void SetPreviewDC(HDC hDC);				// sets a device context for preview pictures

	// *** for alert senders
	virtual int  SendAlert(int alert, int param);	// sent by alerting objects

	virtual void SetTransfer(void *pTransfer);		// must be set by transfer object that received this object
	virtual HDC  GetPreviewDC();					// gets teh device context for preview pictures

protected:
	int (_stdcall *pCallBack)(int alert, int param, void *pTransfer, void *pUser);
	void *pUser;
	int flags;
	void *pTransfer;
	HDC hDC;
};

/***** Transfer Object Section *****/

class DLLEXPORT Transfer
{
public:
	Transfer(void);
	~Transfer(void);

	// transfer functions

	///******** transfers a whole path (with all subdirectories)

	// targetSizeMB: desired size of the target path in megabytes
	// pointer to stopflag: flag usually set to 0, is set to 1 to stop processing and return
	virtual int TransferPath(char *targetPath, char *stopFlag);
				// return codes for TransferPath
				// >0 success - number of MegaBytes written
				//  0 user aborted (stopFlag == 1)
				// -2 if stream is scrambled and either descrambling failed or is disabled
				// -5 source was not opened with "Open"
				// -6 cannot authorize drive
	
	///******* transfer a VOB file

	// targetSizeMB: desired size of the target file in megabytes
	// pointer to stopflag: flag usually set to 0, is set to 1 to stop processing and return
	virtual int TransferFile(char *srcPath, char *targetPath, int targetSizeMB, int tsSrcReadSECT, int tsTargetWrittenSECT, int VTSnum, int VOBnum, char *stopFlag);

	// decryption macros
	#define DISABLE_DECSS				0
	#define ENABLE_DECSS				1

	// authorization macros
	#define DISABLE_AUTHORIZATION		0
	#define ENABLE_AUTHORIZATION		1

	// demo macros
	#define DISABLE_DEMO_MODE			0
	#define ENABLE_DEMO_MODE			1

	// quality control routines
	// 0..fastest, 1..fast, 2..best performance, 3..good quality, 4..best quality
	virtual void SetQuality(int quality);		// sets quality option, numbers outside the range 0..4 will be rounded to the next valid option
	virtual int  GetQuality();					// returns the actual quality option

	// movie only mode, standard is disabled (=custom mode)
	// during movie only mode only the options for the biggest title set are being considered, all other title sets are ignrored
	virtual void SetMovieOnly(bool movieOnly);	// enables (1) or disables (0) movie only mode
	virtual int  GetMovieOnly();				// returns whether movie only mode is enabled (1) or disabled (=custom mode) (0)

	// decryption, standard is disabled
	virtual int  SetUseDeCSS(bool useDeCSS);	// enables (1) or disables (0) decrypting
	virtual int  GetUseDeCSS();					// returns whether decrypting is enabled (1), disabled (0) or not available (-1)

	// authorization, standard is disabled
	// if source is not a dvd drive, no authorization attempt will be made
	virtual int  SetAuthorizeDrive(bool authorizeDrive);	// enables (1) or disables (0) authorization of drive
	virtual int  GetAuthorizeDrive();						// returns whether authorization is enabled (1), disabled (0) or not available (-1)

	// demo functionality
	virtual void SetDemoMode(bool demoMode);				// sets the demo mode (1 = demo, 0 = normal) - can be set interactively while transferring
	virtual int  GetDemoMode();								// returns the current demo mode

	virtual float GetCompressionRatio();// return the actual compression ratio used for the video stream
	virtual int GetReadKB();			// return the kilobytes read
	virtual int GetWrittenKB();			// return the kilobytes written
	virtual int GetNumFrames();			// return the (absolute) number of frames processed
	virtual int GetReadLimitKB();		// return the limit of kilobytes that will be read by the codec
										//   this is the sizeof(dvd) - sizeof(fixed, nonreencoded files)
	virtual int GetReadBufferOccupancyKB();  // return the kilobytes in the read  buffer
	virtual int GetWriteBufferOccupancyKB(); // return the kilobytes in the write buffer

	virtual char *GetActualFileName();	// return the actual file name being processed
	virtual int GetElapsedTimeSec();	// return the seconds elapsed since deencoder was started
	virtual int GetEstimatedTimeSec();  // return the seconds of how long the encoding process will last
	virtual int GetRemainingTimeSec();  // return the seconds till the end of encoding process

	virtual int Open(char *srcPath);	// opens a video path and retrieves information about it
				// return codes for Open
				//  1 success
				// -2 if stream is scrambled and either descrambling failed or is disabled
				// -4 if there is a read problem or source directory is empty

	virtual int SetTargetSizeMB(int targetSize);						// sets the desired target size in megabytes
	virtual int GetTargetSizeMB();										// return the target size in megabytes
	virtual int GetSourceSizeMB(bool movieOnly);						// return the size of 
																		//   - the whole source DVD (movieOnly = false) 
																	    //	 - the main movie (movieOnly = true)
	virtual int ForceTargetSize(bool force);							// forces the target size to be 99% (1) 
																		// instead of 95% (0) of the target size
																		// returns the former force status
	virtual int GetNumVTS();											// return the number of video title sets
	virtual int GetVTSSize(int VTSIndex);								// return the size of one VTS

	virtual int GetNumVTSLang(int VTSIndex); // return the number of languages in a VTS
	virtual int GetVTSLangDesc(int VTSIndex, int langIndex, char *buffer, int szBuffer); // return the description of a language in a VTS
	virtual int GetVTSLangSelected(int VTSIndex, int langIndex);						 // return 1 if language in VTS is selected
	virtual int GetVTSLangDefault(int VTSIndex, int langIndex);							 // return index of standard language in a VTS (movie only)
	virtual int SetVTSLangSelected(int VTSIndex, int langIndex, int selected);			 // selects a language in a VTS (1 - selected, 0 not selected)
	virtual int SetVTSLangDefault(int VTSIndex, int langIndex);							 // selects standard language in a VTS, -1 for no language  (movie only)
				// return codes for SetVTSLangSelected and SetVTSLangDefault
				//  2 warning: success but will delete standard language (set proper standard language first)
				//  1 success
				// -1 parameters (indices) out of range
				// -5 source was not opened with "Open"
				// return codes specific for SetVTSLangDefault
				//  2 warning: success but standard language not selected for inclusion (language will be deleted)

	virtual int GetNumVTSSub(int VTSIndex);  // return the number of subpicture streams in a VTS
	virtual int GetVTSSubDesc(int VTSIndex, int subIndex, char *buffer, int szBuffer); 	 // return the description of a subpicture stream in a VTS
	virtual int GetVTSSubSelected(int VTSIndex, int subIndex);							 // return 1 if subpicture stream in VTS is selected
	virtual int GetVTSSubDefault(int VTSIndex, int subIndex);							 // return index of standard subpicture stream in a VTS or -9 for no subpicture stream (movie only)
	virtual int SetVTSSubSelected(int VTSIndex, int subIndex, int selected);			 // selects a subpicture stream in a VTS (1 - selected, 0 not selected)
	virtual int SetVTSSubDefault(int VTSIndex, int subIndex);							 // selects standard subpicture stream in a VTS, -1 for no subpicture stream (movie only)
				// return codes for SetVTSSubSelected and SetVTSSubDefault
				//  2 warning: success but will delete standard subpicture (set proper standard subpicture first)
				//  1 success
				// -1 parameters (indices) out of range
				// -5 source was not opened with "Open"
				// return codes specific for SetVTSSubDefault
				//  2 warning: success but standard subpicture not selected for inclusion (subpicture will be deleted)

	// for C++ languages only
	virtual void SetAlert(TransferAlert *alert);										 // sets a pointer to an alert object
																						 // the alert object can be changed at will before and after setting it in transfer

protected:
	virtual __int64 EvaluatePath(char *path);
	virtual void    CopyAllFiles     (char *srcPath, char *targetPath);
	virtual int TransferFiles        (char *srcPath, char *targetPath, /*int *srcReadMB, int *targetWrittenMB, */int srcSizeMB, int targetSizeMB, char *stopFlag);
	virtual int SearchEncryption     (char *srcPath, char *stopFlag);
	virtual int GenerateIgnoreTable(int VTSnum, int VOBnum);
	int GetFileSizeSECT(char *srcPath);
	virtual int GetVOBSizeMB(char *srcPath, int VTSnum, int VOBnum);
	virtual int GetVOBSizeSECT(char *srcPath, int VTSnum, int VOBnum);
	virtual int GetMovieOnlyVTSNum();
	virtual int GetMovieOnlySizeMB();

	int globalReadKB, globalWrittenKB, globalNumFrames;
	int localReadKB, localWrittenKB, localNumFrames;
	int readBufferOccupancy, writeBufferOccupancy;
	float compressionRatio;
	char actualFileName[256];
	char srcPath[256];
	int opened;
	int quality;
	int targetSizeMB;
	int sourceSizeMB;
	int sourceSizeKB;
	int forceTargetSize;
	int demoMode;

	int baseTimeMSec, elapsedTimeMSec, estimatedTimeMSec, remainingTimeMSec;
	float savedEstimatedTimeMSec;

	int defAud, defSub;
	int movieOnly;
	int useDeCSS, authorize;

	void *IgnoreTable;
	char ***SelectTable;

	void *openIFO;
	void *mainIFO;
	void *currIFO;

	void *CSS;

	TransferAlert *alert;
};

/*** plain C exports ***/
// TransferAlert Object
#ifdef __cplusplus
extern "C" 
{
#endif

	void _stdcall SetCallBack(int (_stdcall *pCallBack)(int alert, int param, void *pTransfer, void *pUser), void *pUser);

	void _stdcall AddAlert(int alert);
	void _stdcall SubAlert(int alert);
	bool _stdcall GetAlert(int alert);
	int  _stdcall GetAllAlert();
	void _stdcall SubAllAlert();

	void _stdcall SetPreviewDC(HDC hDC);

#ifdef __cplusplus
}
#endif

// Transfer Object
#ifdef __cplusplus
extern "C" 
{
#endif

	int _stdcall TransferPath(char *targetPath, char *stopFlag);

	void _stdcall SetQuality(int quality);
	int  _stdcall GetQuality();

	void _stdcall SetMovieOnly(bool movieOnly);
	int  _stdcall GetMovieOnly();

	int  _stdcall SetUseDeCSS(bool useDeCSS);
	int  _stdcall GetUseDeCSS();

	int  _stdcall SetAuthorizeDrive(bool authorizeDrive);
	int  _stdcall GetAuthorizeDrive();

	void _stdcall SetDemoMode(bool demoMode);
	int  _stdcall GetDemoMode();

	float _stdcall GetCompressionRatio();
	int _stdcall GetReadKB();
	int _stdcall GetWrittenKB();
	int _stdcall GetNumFrames();
	int _stdcall GetReadLimitKB();
	int _stdcall GetReadBufferOccupancyKB();
	int _stdcall GetWriteBufferOccupancyKB();

	char * _stdcall GetActualFileName();
	int _stdcall GetElapsedTimeSec();
	int _stdcall GetEstimatedTimeSec();
	int _stdcall GetRemainingTimeSec();

	int _stdcall Open(char *srcPath);
	int _stdcall SetTargetSizeMB(int targetSize);
	int _stdcall GetTargetSizeMB();
	int _stdcall GetSourceSizeMB(bool movieOnly);
	int _stdcall ForceTargetSize(bool force);
	int _stdcall GetNumVTS();
	int _stdcall GetVTSSize(int VTSIndex);

	int _stdcall GetNumVTSLang(int VTSIndex);
	int _stdcall GetVTSLangDesc(int VTSIndex, int langIndex, char *buffer, int szBuffer);
	int _stdcall GetVTSLangSelected(int VTSIndex, int langIndex);
	int _stdcall GetVTSLangDefault(int VTSIndex, int langIndex);
	int _stdcall SetVTSLangSelected(int VTSIndex, int langIndex, int selected);
	int _stdcall SetVTSLangDefault(int VTSIndex, int langIndex);

	int _stdcall GetNumVTSSub(int VTSIndex);
	int _stdcall GetVTSSubDesc(int VTSIndex, int subIndex, char *buffer, int szBuffer);
	int _stdcall GetVTSSubSelected(int VTSIndex, int subIndex);
	int _stdcall GetVTSSubDefault(int VTSIndex, int subIndex);
	int _stdcall SetVTSSubSelected(int VTSIndex, int subIndex, int selected);
	int _stdcall SetVTSSubDefault(int VTSIndex, int subIndex);

#ifdef __cplusplus
}
#endif
