/*****************************************************************************
 * CompressionStrategy.cpp: The Miracle behind ShrinkTo5
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "StdAfx.h"
#include "compressionstrategy.h"

CompressionStrategy::CompressionStrategy(void)
{
	maxNumPics = 18;
	maxNumSlices = 18 * 48;
	currNumPics = 0;
	currNumSlices = 0;

	picSize = new PICTURE_SIZE[maxNumPics];
	sliSize = new SLICE_SIZE[maxNumSlices];
}

CompressionStrategy::~CompressionStrategy(void)
{
	delete picSize;
	delete sliSize;
}

void CompressionStrategy::CreateStrategy(STARTCODE_POS *posTable, int seqSize, int numIPictures, int numPPictures, int numBPictures, int skipAmount)
{
	this->posTable = posTable;
	this->seqSize = seqSize;
	this->skipAmount = skipAmount;

	this->numIPictures = numIPictures;
	this->numPPictures = numPPictures;
	this->numBPictures = numBPictures;

	numPictures = numIPictures + numPPictures + numBPictures;

	ReAllocMemory();
	AssignLengths();
	MakeStrategy();
}

void CompressionStrategy::AssignLengths()
{
	currNumPics = 0;
	currNumSlices = 0;

	int currSlices = 0;
	int effPicSize = 0;
	effSeqSize = 0;

	sizeIPictures = 0;
	sizePPictures = 0;
	sizeBPictures = 0;

	int wasPrevSlice = 0;
	int lastPos = 0;
	for(int i = 0; i < posTable->numEntries; i++)
	{
		int type = posTable->pEntries[i].entryType;
		int currSize = posTable->pEntries[i].pos - lastPos;

		// if the former type was a slice
		if(wasPrevSlice)
		{
			// save slice size
			sliSize[currNumSlices].size = currSize;
			// rise picture effective size
			effPicSize += currSize;

			currNumSlices++;
			currSlices++;
		}

		wasPrevSlice = 0;

		// if we have now a picture
		if(type == I_PICTURE || type == P_PICTURE || type == B_PICTURE)
		{
			picSize[currNumPics].sliceIdx = currNumSlices;			// save the actual starting slice index
			picSize[currNumPics].pictureType = type;				// save picture type

			// if we have had pictures before
			if(currNumPics > 0)
			{
				picSize[currNumPics - 1].numSlices = currSlices;	// save number of slices for this picture
				picSize[currNumPics - 1].size = effPicSize;			// save effective picture size

				// add category size
				if(picSize[currNumPics - 1].pictureType == I_PICTURE) sizeIPictures += effPicSize;
				if(picSize[currNumPics - 1].pictureType == P_PICTURE) sizePPictures += effPicSize;
				if(picSize[currNumPics - 1].pictureType == B_PICTURE) sizeBPictures += effPicSize;

				effSeqSize += effPicSize;
				effPicSize = 0;
				currSlices = 0;
			}

			currNumPics++;
		}

		// if we have now a slice
		if(type == SLICE)
		{
			sliSize[currNumSlices].refIdx = i;						// save reference index
			wasPrevSlice = 1;
		}

		lastPos = posTable->pEntries[i].pos;
	}

	// if the former type was a slice
	if(wasPrevSlice)
	{
		// calc last size
		int currSize = seqSize - lastPos;

		// save slice size
		sliSize[currNumSlices].size = currSize;
		// rise picture effective size
		effPicSize += currSize;

		currNumSlices++;
		currSlices++;
	}

	// if we have had pictures before
	if(currNumPics > 0)
	{
		picSize[currNumPics - 1].numSlices = currSlices;			// save number of slices for this picture
		picSize[currNumPics - 1].size = effPicSize;					// save effective picture size

		// add category size
		if(picSize[currNumPics - 1].pictureType == I_PICTURE) sizeIPictures += effPicSize;
		if(picSize[currNumPics - 1].pictureType == P_PICTURE) sizePPictures += effPicSize;
		if(picSize[currNumPics - 1].pictureType == B_PICTURE) sizeBPictures += effPicSize;

		effSeqSize += effPicSize;
	}
}

void CompressionStrategy::ReAllocMemory()
{
	// if we need more memory for pictures
	if(numPictures > maxNumPics)
	{
		maxNumPics = numPictures * 2;

		delete picSize;
		picSize = new PICTURE_SIZE[maxNumPics];
	}

	// if we need more memory for slices
	if(posTable->maxNumEntries > maxNumSlices)
	{
		maxNumSlices = posTable->maxNumEntries;

		delete sliSize;
		sliSize = new SLICE_SIZE[maxNumSlices];
	}
}

void CompressionStrategy::MakeStrategy()
{
	// calc new skip amount
	skipAmount = ((__int64)effSeqSize * (__int64)skipAmount) / (__int64)seqSize;
	skipAmount = min(effSeqSize, skipAmount);

	// calc taking amount
	int takeAmount = effSeqSize - skipAmount;

	int takeIAmount = 0;
	int takePAmount = 0;
	int takeBAmount = 0;

	// do affect I pictures the least
	if(sizeBPictures + sizePPictures < takeAmount) 
	{
		takeIAmount = sizeIPictures - skipAmount;
		takePAmount = sizePPictures;
		takeBAmount = sizeBPictures;
	}
	// rescue as much P pictures as possible
	else if(sizeBPictures < takeAmount)
	{
		takePAmount = sizePPictures - (skipAmount - sizeIPictures);
		takeBAmount = sizeBPictures;
	}
	// last but not least rescue as much of the B pictures as possible
	else 
	{
		takeBAmount = sizeBPictures - (skipAmount - sizeIPictures - sizePPictures);
	}

	// calc slice skippings for each category
	/*if(takeIAmount != 0)*/ CalcSkips(I_PICTURE, sizeIPictures, takeIAmount);
	/*if(takePAmount != 0)*/ CalcSkips(P_PICTURE, sizePPictures, takePAmount);
	/*if(takeBAmount != 0)*/ CalcSkips(B_PICTURE, sizeBPictures, takeBAmount);
}

int __cdecl compareSlices(const void *sli1, const void *sli2)
{
	SLICE_SIZE *slice1 = (SLICE_SIZE *)sli1;
	SLICE_SIZE *slice2 = (SLICE_SIZE *)sli2;

	if(slice1->size >  slice2->size) return 1;
	if(slice1->size == slice2->size) return 0;
									 return -1;
}

void CompressionStrategy::CalcSkips(int pictureType, int sizePictures, int takeAmount)
{
	for(int i = 0; i < currNumPics; i++)
	{
		// if we have found our picture
		if(picSize[i].pictureType == pictureType)
		{
			int takeLocalAmount = ((__int64)takeAmount * (__int64)picSize[i].size) / (__int64)sizePictures; 
			    takeLocalAmount = min(picSize[i].size, takeLocalAmount);

			// if we should skip the whole picture
			if(takeLocalAmount == 0)
			{
				int sliceIndex = picSize[i].sliceIdx;
				// skip all sliced
				while((sliceIndex - picSize[i].sliceIdx) < picSize[i].numSlices)
				{
					// set skip flag
					posTable->pEntries[sliSize[sliceIndex].refIdx].skip = 1;
					sliceIndex++;
				}
			}
			// if we do not have to take the whole picture
			else if(takeLocalAmount != picSize[i].size)
			{
				int localSkip = picSize[i].size - takeLocalAmount;
				int localSkipped = 0;
				int localTaken = 0;

				// sort the picture
				qsort(&sliSize[picSize[i].sliceIdx], picSize[i].numSlices, sizeof(SLICE_SIZE), compareSlices);

				int sliceIndex = picSize[i].sliceIdx;
				// skip small to middle sized slices
				while((localSkip - sliSize[sliceIndex].size - localSkipped) >= 0 &&
					  (sliceIndex - picSize[i].sliceIdx) < picSize[i].numSlices)
				{
					// count local skipped bytes
					localSkipped += sliSize[sliceIndex].size;
					// set skip flag
					posTable->pEntries[sliSize[sliceIndex].refIdx].skip = 1;
					sliceIndex++;
				}
			}
		}
	}
}
