/*****************************************************************************
 * IFO.h: Analyse IFO Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#pragma once

#include "IFOEdit.h"

#undef SECTOR_SIZE

#define REFERENCE_IFO		1
#define REFERENCE_MENU_VOB  2
#define REFERENCE_TITLE_VOB	3

#define MENU_VOB			1
#define TITLE_VOB			0	

#define STREAM_AUDIO		0
#define STREAM_SUBSTREAM	1

typedef struct tagSECTOR_LUT
{
	int srcSector;
	int targetSector;
	int reference;
	int userData;
} SECTOR_LUT;

typedef struct tagAVTABLE
{
	unsigned char audioAvailable[8];
	unsigned char audioStreamType[8];
	unsigned char audioStreamNum[8];
	unsigned char videoAvailable[32];
	unsigned char videoStreamNum[128];
} AVTABLE;

class IFO
{
public:
	IFO(void);
	~IFO(void);

	// reads original IFO/BUP and creates the sector lookup table
	virtual int Load(char *filePath);
	// writes modified IFO/BUP from the sector lookup table
	virtual int Save(char *filePath);
	// close file and leave unmodified
	virtual int Close();
	// returns the internal buffer
	virtual unsigned char *GetIFOBuffer();
	// set movie only mode
	virtual void SetMovieOnly(bool movieOnly);

	// manipulation routines
	virtual int SearchMovieTitle(int *PGCidx);					// searches for movie in IFO in movie only mode according to a heuristic
	virtual void MakeMovieTitlePlayable(int TT, int PGCidx, int defAud, int defSub, char *selAudFlags, char *selSubFlags);	
																// makes movie title playable (selects proper language and subtitle, makes PGC 
																// an entry one, removes prohibited user ops)
	virtual void SelectMovieTitle(int TT, int VTS, unsigned char *orgIFOBuffer);	// selects movie in VIDEO_TS.IFO in movie only mode

	virtual void CopyAudioVideoAttributes(unsigned char *srcIFOBuffer); // copies the audio and video attributes into a 
																		// VIDEO_TS (video manager) from a VIDEO_01_0.IFO (video title set). 
																		// only for movie only mode

	virtual void SetFirstMenuVOBSector(int sector);				// set begining of the menu VOB
	virtual void SetFirstTitleVOBSector(int sector);			// set beginning of the title VOB
	virtual void SetLastBUPSector(int sector);					// set end of title set
	virtual int VMG_SetTitleStartSector(int title, int sector);	// set start of earch title set (first sector of each IFO)

	virtual int  GetNextSectorEntry(int reference);					// returns the next entry in the sector LUT
	virtual void SetSectorEntry(int srcSector, int targetSector, int reference);	// sets the target sector in the corresponding sector LUT entry
	virtual void SetCurrSectorEntry(int srcSector, int reference);	// initializes the get-next-sector-entry iterator

	// information routines (for VMGs IFOs only)
	virtual int GetNumVTS();															 // return the number of video title sets
	virtual int GetNumVTSLang(int VTSIndex); // return the number of languages in a VTS
	virtual int GetVTSLangDesc(int VTSIndex, int langIndex, char *buffer, int szBuffer); // return the description of a language in a VTS
	virtual int IsVTSLangMenuItem(int VTSIndex, int langIndex);							 // return whether language is part of a titleset menu vob
	virtual int GetNumVTSSub(int VTSIndex);  // return the number of subpicture streams in a VTS
	virtual int GetVTSSubDesc(int VTSIndex, int subIndex, char *buffer, int szBuffer); 	 // return the description of a subpicture stream in a VTS
	virtual int IsVTSSubMenuItem(int VTSIndex, int subIndex);							 // return whether subpicture stream is part of a titleset menu vob

	// information routines for streams
	virtual int GetPGCMarks(int reference, int **markSECT);			// returns the beginning marks of a PGC
	virtual int GetAVTable(int reference, AVTABLE **AVStreamID);	// returns the audio/video (sub)stream numbering tables for each PGC
	virtual int GetNumSub(int reference);							// returns the number of subs for the menu or title VOB
	virtual int GetNumLang(int reference);							// returns the number of language for the menu or title VOB
	virtual int GetLangCompression(int reference, int langIndex);	// returns the compression algorithm used for the audio stream


	// see if IFO is a VMG or VTS
	virtual int isVTS() { return !isVMG/*return ifoEdit->isVTS()*/; }

	SECTOR_LUT *sectorLUT;
	int sectorLUTEntries;

protected:
	virtual void ProcessVMGSectorLUTs(bool write);	// processes the IFO structure and creates the sector lookup tables (for VIDEO_TS.IFO)
	virtual void ProcessVTSSectorLUTs(bool write);	// processes the IFO structure and creates the sector lookup tables (for VTS_xx_x.IFO)

	virtual int ProcessC_ADT(int pC_ADT, int LUT, int isMenu, bool write);				// processes C_ADT structures
	virtual int ProcessVOBU_ADMAP(int pVOBU_ADMAP, int LUT, int isMenu, bool write);	// processes VOBU_ADMAP structures
	virtual int ProcessPGCI_UT(int pPGCI_UT, int LUT, bool write);						// processes PGCI_UT structures
	virtual int ProcessPGCI_LU(int pPGCI_LU, int LUT, int isMenu, bool write);			// processes PGCI/LU structures

	virtual int GetVTS_IFOPointer(int VTSIndex);										// returns a pointer to the VTS IFO in the VMG IFO

	//CIFOEdit *ifoEdit;
	int isVMG;

	SECTOR_LUT titleStartLUT[128];
	int titleStartLUTEntries;

	int MenuVOBStartSECT;
	int TitleVOBStartSECT;

	int currLUTVOBEntry;
	int currLUTIFOEntry;

	int lastBUPSector;
	int *markSECT;
	AVTABLE *AVStreamID;

	unsigned char *IFOBuffer;
	int IFOBufferSize, currIFOBufferSize;

	int movieOnly;
};



