/*****************************************************************************
 * IOLayer.cpp: Custom IOLayer Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "StdAfx.h"
#include "iolayer.h"
#include <process.h>

// working thread - sleeps more often than not
unsigned __stdcall WorkingThread(void *pIOLayer);

IOLayer::IOLayer(void)
{
	kBThroughput = 0;
	SECTOR_SIZE = 2048;
//	MAX_SECTORS = 4096;
	MAX_SECTORS = 2048;
//	MAX_SECTORS = 1024;
	MAX_BUFFERS = 4;
	SLEEP_TIME_MS = 10;

	ioElement = new IO_BUFFER *[MAX_BUFFERS];
	for(int i = 0; i < MAX_BUFFERS; i++)
	{
		ioElement[i] = new IO_BUFFER;
	}

	hFile = NULL;
	threadHandle = 0;
	opened = 0;
}

IOLayer::~IOLayer(void)
{
	if(opened == 1)	Close();

	for(int i = 0; i < MAX_BUFFERS; i++)
	{
		delete ioElement[i];
	}
	delete ioElement;
}

// set io buffer size in sectors
void IOLayer::SetBufferSize(int numSectors)
{
	MAX_SECTORS = numSectors;
}

// sector size in bytes
void IOLayer::SetSectorSize(int numBytes)
{
	SECTOR_SIZE = numBytes;
}

// open a file for async io and init buffers
int IOLayer::Open(unsigned char *file, int rwFlag)
{
	int IOFlag, ShareFlag, DispFlag, FileFlag;
	int i;

	this->rwFlag = rwFlag;

	// if we are on windows nt then read overlapped
	FileFlag = FILE_FLAG_NO_BUFFERING;

	// readonly
	if(!rwFlag)
	{
		IOFlag = GENERIC_READ;
		ShareFlag = FILE_SHARE_READ;
		DispFlag = OPEN_EXISTING;
	}
	// writeonly
	else
	{
		IOFlag = GENERIC_WRITE;
		ShareFlag = 0;
		DispFlag = CREATE_ALWAYS;
	}

	// open file
	hFile = CreateFile((LPCSTR)file, IOFlag, ShareFlag, NULL, DispFlag, 
					   FileFlag, NULL);

	// if an error occurs
	if(hFile == INVALID_HANDLE_VALUE)
	{
		LPVOID lpMsgBuf;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,	GetLastError(),	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf, 0, NULL);
		MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );
		hFile = NULL;
		return -1;
	}

	// mark buffers as empty
	// allocate buffers (!!! if working set is too small this functions will fail !!!)
	for(i = 0; i < MAX_BUFFERS; i++)
	{
		ioElement[i]->maxCurrSector = 0;

		ioElement[i]->pBuffer = (unsigned char *) VirtualAlloc(NULL, MAX_SECTORS * SECTOR_SIZE, MEM_COMMIT, PAGE_READWRITE);
		VirtualLock(ioElement[i]->pBuffer, MAX_SECTORS * SECTOR_SIZE);
	}

	// intialize parameters
	currSector = 0;
	currOffset = 0;

	unsigned int lowPart, highPart;
	lowPart = GetFileSize(hFile, (LPDWORD)&highPart);
	FileSize = (__int64)lowPart + ((__int64)highPart << 32);
	CurrPos = 0;

	InitializeCriticalSection(&csection);

	// if opened for reading
	if(!rwFlag)
	{
		// fill buffers
		for(int i = 0; i < MAX_BUFFERS; i++)
		{
			internalRead(ioElement[i]);
		}
	}

	int id;
	stopFlag = 0;
	threadHandle = _beginthreadex(NULL, 6553600, WorkingThread, this, 0, (unsigned int *) &id);

	opened = 1;

	return 0;
}

// close the file for async io and free buffers
void IOLayer::Close()
{
	// if not already closed
	if(!opened) return;

	if(hFile)
	{
		// let our thread exit
		stopFlag = 1;
		do { Sleep(SLEEP_TIME_MS); } while(threadHandle != 0);

		// write the rest of the buffer
		if(rwFlag) 
		{
			IO_BUFFER *tioElement;
			do
			{
				// clear io element
				tioElement = NULL;
				// search for buffers to flush to disk
				if(ioElement[0]->maxCurrSector > 0)
				{
					tioElement = ioElement[0];
				}
				else
				{
					for(int i = MAX_BUFFERS - 1; i > 0; i--)
					{
						if(ioElement[i]->maxCurrSector > 0)
						{
							tioElement = ioElement[i];
							break;
						}				
					}
				}
				// if there is really a buffer to flush to disk
				if(tioElement != NULL) internalWrite(tioElement);
			} while(tioElement != NULL);

			// if there is a non full buffer left (the first one)
			if(currSector > 0) 
			{
				ioElement[0]->maxCurrSector = currSector;
				internalWrite(ioElement[0]);

				currSector = 0;
			}
		}

		// close file
		CloseHandle(hFile);
		hFile = NULL;
	
		// free buffers
		for(int i = 0; i < MAX_BUFFERS; i++)
		{
			VirtualUnlock(ioElement[i]->pBuffer, MAX_SECTORS * SECTOR_SIZE);
			VirtualFree(ioElement[i]->pBuffer, 0, MEM_RELEASE);
		}
	}

	DeleteCriticalSection(&csection);

	opened = 0;
}

// read from buffer
int IOLayer::Read(void *buffer)
{
	// if the buffer has enough data
	if(currSector < ioElement[0]->maxCurrSector)
	{
		CopyMemory(buffer, &(ioElement[0]->pBuffer[currSector * SECTOR_SIZE]), SECTOR_SIZE);
		currSector++;
		currOffset += SECTOR_SIZE;
		return currOffset;
	}
	// if buffer is empty
	else
	{
		EnterCriticalSection(&csection);
		ioElement[0]->maxCurrSector = 0;

		for(int i = 1; i < MAX_BUFFERS; i++)
		{
			IO_BUFFER *tioElement = ioElement[i - 1];
			ioElement[i - 1] = ioElement[i];
			ioElement[i] = tioElement;
		}
		LeaveCriticalSection(&csection);

		if(FileSize == CurrPos && ioElement[0]->maxCurrSector == 0) return -1;
		while(ioElement[0]->maxCurrSector == 0)	
		{
			Sleep(SLEEP_TIME_MS);
		}
		currSector = 0;

		return Read(buffer);
	}
}

// read directly from file
int IOLayer::Read(void *buffer, __int64 offset)
{
	// if sector is beyond file size return error
	if(FileSize == offset) return -1;

	// initialize read
	int BytesToRead = SECTOR_SIZE;
	int BytesRead = 0;

	int currPosHigh = offset >> 32;
	int currPos = SetFilePointer(hFile, offset & 0xFFFFFFFF, (PLONG)&currPosHigh, FILE_BEGIN);
	int retVal = ReadFile(hFile, buffer, BytesToRead, (LPDWORD)&BytesRead, NULL);
	SetFilePointer(hFile, currPos, (PLONG)&currPosHigh, FILE_BEGIN);

	// if an unexpected error occurs
	if(!retVal && GetLastError() != ERROR_IO_PENDING)
	{
		LPVOID lpMsgBuf;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,	GetLastError(),	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf, 0, NULL);
		MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );
		return -1;
	}

	return SECTOR_SIZE;
}

// read from file into buffer
int IOLayer::internalRead(IO_BUFFER *ioElement)
{
		// if there is nothing to read
		if(FileSize == CurrPos)	return -1;

		// initialize read
		int BytesToRead = min(FileSize - CurrPos, (__int64)(MAX_SECTORS * SECTOR_SIZE));

		// int retVal = ReadFile(hFile, pBuffer[1], BytesToRead, NULL, &ovStruct);
		DWORD BytesRead;
		int retVal = ReadFile(hFile, ioElement->pBuffer, BytesToRead, &BytesRead, NULL);

		// if an unexpected error occurs
		if(!retVal && GetLastError() != ERROR_IO_PENDING)
		{
			LPVOID lpMsgBuf;
			FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
						  NULL,	GetLastError(),	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
						  (LPTSTR) &lpMsgBuf, 0, NULL);
			MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
			LocalFree( lpMsgBuf );
			return -1;
		}

		// init parameters
		ioElement->maxCurrSector = BytesToRead / SECTOR_SIZE;

		// advance position
		CurrPos += BytesToRead;
		kBThroughput += BytesToRead;

		return 0;
}

// write into buffer
int IOLayer::Write(void *buffer)
{
	// if the buffer has enough data
	if(currSector < MAX_SECTORS)
	{
		CopyMemory(&(ioElement[0]->pBuffer[currSector * SECTOR_SIZE]), buffer, SECTOR_SIZE);
		currSector++;
		currOffset += SECTOR_SIZE;
		return currOffset;
	}
	// if buffer is empty
	else
	{
		EnterCriticalSection(&csection);
		ioElement[0]->maxCurrSector = currSector;

		for(int i = MAX_BUFFERS - 1; i > 0; i--)
		{
			IO_BUFFER *tioElement = ioElement[i];
			ioElement[i] = ioElement[i - 1];
			ioElement[i - 1] = tioElement;
		}
		LeaveCriticalSection(&csection);

		while(ioElement[0]->maxCurrSector > 0) 
		{
			Sleep(SLEEP_TIME_MS);
		}
		currSector = 0;

		return Write(buffer);
	}
}

// write from buffer into file
int IOLayer::internalWrite(IO_BUFFER *ioElement)
{
	// initialize write
	int BytesToWrite = ioElement->maxCurrSector * SECTOR_SIZE;

	DWORD BytesWritten;
	int retVal = WriteFile(hFile, ioElement->pBuffer, BytesToWrite, &BytesWritten, NULL);

	// if an unexpected error occurs
	if(!retVal && GetLastError() != ERROR_IO_PENDING)
	{
		LPVOID lpMsgBuf;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,	GetLastError(),	MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
						(LPTSTR) &lpMsgBuf, 0, NULL);
		MessageBox( NULL, (LPCTSTR)lpMsgBuf, "Error", MB_OK | MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );
		return -1;
	}

	// advance position
	CurrPos += BytesToWrite;
	kBThroughput += BytesToWrite;

	// init parameters
	ioElement->maxCurrSector = 0;

	return 0;
}

unsigned __stdcall WorkingThread(void *pIOLayer)
{

	IOLayer *pIO = (IOLayer *) pIOLayer;

	// do until thread is stopped
	do
	{
		IO_BUFFER *tioElement = NULL;
		// *** ENTER critical section
		EnterCriticalSection(&pIO->csection);
		// if thread is for reading
		if(pIO->rwFlag == 0)
		{
			for(int i = 0; i < pIO->MAX_BUFFERS; i++)
			{
				if(pIO->ioElement[i]->maxCurrSector == 0 && pIO->CurrPos != pIO->FileSize)
				{
					tioElement = pIO->ioElement[i];
					break;
				}				
			}
		}
		// if thread is for writing
		else
		{
			if(pIO->ioElement[0]->maxCurrSector > 0)
			{
				tioElement = pIO->ioElement[0];
			}
			else
			{
				for(int i = pIO->MAX_BUFFERS - 1; i > 0; i--)
				{
					if(pIO->ioElement[i]->maxCurrSector > 0)
					{
						tioElement = pIO->ioElement[i];
						break;
					}				
				}
			}
		}
		LeaveCriticalSection(&pIO->csection);
		// *** LEAVE critical section

		if(tioElement == NULL) Sleep(pIO->SLEEP_TIME_MS);
		else
		{
			if(pIO->rwFlag == 0) pIO->internalRead(tioElement);
			else				 pIO->internalWrite(tioElement);
		}
	} while(pIO->stopFlag == 0);

	pIO->threadHandle = 0;

	return 0;
}