/*****************************************************************************
 * PESPacket.cpp: PESPacket Buffer Class
 *****************************************************************************
 * Copyright (C) 2004-2005 ShrinkTo5
 *
 * Authors: Ocrana		 <info@shrinkto5.com>
 *          Andrej Grecu <info@shrinkto5.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include "StdAfx.h"
#include "pespacket.h"

PESPacket::PESPacket(void)
{
}

PESPacket::PESPacket(unsigned char *buffer, int length)
{
	AttachBuffer(buffer, length);
}

PESPacket::~PESPacket(void)
{
}

// attach a buffer to the object
// read the contents of the packet header
int PESPacket::AttachBuffer(unsigned char *buffer, int length)
{
	pBuffer = buffer;
	bufferSize = length;

	// init bitstream with new buffer
	bstream.AttachBuffer(pBuffer, bufferSize);

	// **** ---> see ISO 13818-1 Table 2-18 (PES-Packet) <--- ***
	// **** ---> see ISO 13818-1 Table 2-19 (Stream ID Assignments) <--- ***
	// **** ---> see ISO 13818-2 Table 6-1  (Start Code Values) <--- ***

	int val;
	// jump start code
	val = bstream.ReadBits(24);
	_ASSERT(val == 0x000001);
	// get stream id
	streamID = bstream.ReadBits(8);
	// if PES packet is a video stream (codes 1110 xxxx) or a private stream 1
	if((streamID >= 0xE0 && streamID <= 0xEF) || streamID == 0xBD)
	{
		// packet length is 6 bytes longer than in packetheader (starts from the beginning of packet)
		packetSize = bstream.ReadBits(16) + 6;
		// not needed now
		bstream.Seek(2);
		// if packet is scrambled and no DeCSS object is available return with error
		int scrambling = bstream.ReadBits(2);
		if(scrambling != 0) return -2;
		bstream.Seek(1);
		// no aligment flag should be set
		val = bstream.ReadBits(1);
		_ASSERT(val == 0);
		// not needed now		
		bstream.Seek(8);
		// read CRC flag
		int CRC = bstream.ReadBits(1);
		// assume that it is not set
		_ASSERT(CRC == 0);
		// not needed now
		bstream.Seek(1);
		// payload is 72 (9 byte) + PES_data_length bytes from PES packet start
		int header_length = bstream.ReadBits(8);
		_ASSERT(header_length >= 0);
		payLoad = &pBuffer[9 + header_length];
		// calculate payload size
		// if payload does not occupy the entire pack then it is stuffed with 1s
		payLoadSize = packetSize - 9 - header_length;
		// calculate maximum payload size (without stuffing)
		maxPayLoadSize = bufferSize - 9 - header_length;

		#ifdef _DEBUG
			// assume rest of pack is a padding stream
			// rest of pack = bufferSize - packetSize
			if(bufferSize > packetSize/* && streamID != 0xBD*/)
			// if there is a spare space
			{
				// check padding stream
				_ASSERT(pBuffer[packetSize] == 0x00);
				_ASSERT(pBuffer[packetSize + 1] == 0x00);
				_ASSERT(pBuffer[packetSize + 2] == 0x01);
				_ASSERT(pBuffer[packetSize + 3] == 0xBE);
			}
			// assume rest of pack is stuffed with 1s
			// rest of pack = bufferSize - packetSize
/*			if(streamID != 0xBD)
			{*/
				int i;
				for(i = bufferSize - 1; i >= packetSize + 6; i--)
				{
					_ASSERT(pBuffer[i] == 0xFF);
				}
//			}
		/*	// assume sequence header is only found at the beginning of a packet
			// (this test has loopholes at the end of the packet)
			for(i = 9 + header_length + 1; i < packetSize - 3; i++)
			{
				_ASSERT(!(pBuffer[i    ] == 0x00 && 
					      pBuffer[i + 1] == 0x00 && 
						  pBuffer[i + 2] == 0x01 && 
						  pBuffer[i + 3] == 0xB3));
			}*/
			// assume header length 0 (or number of stuffing bytes) for packets without sequence header
/*			if(!(payLoad[0] == 0x00 && 
				 payLoad[1] == 0x00 && 
				 payLoad[2] == 0x01 && 
				 payLoad[3] == 0xB3))
			{
				i = 0;
				int c = 0;
				for(i = -header_length; i < 0; i++)
				{
					if(payLoad[i] == 0xFF) c++;
				}
				_ASSERT(header_length == c);
			}*/
		#endif
	}
	// if PES packet is not a video stream
	else
	{
		payLoad = &pBuffer[4];
		payLoadSize = bufferSize - 4;
		maxPayLoadSize = payLoadSize;
	}

	// if the stream is a private stream then the first byte of the payload is the substream number
	if(streamID == 0xBD)
	{
		subStreamID = payLoad[0];
		payLoad++; // beware for 0xBF there need not be a substream ID in the first payload dword (but it is on DVDs)
	}
	else if(streamID == 0xBB) // system header 0xBB means NAVPack on DVDs
	{
		// assume this can be only a NAVPack
		_ASSERT(payLoad[0x26 - 0x0E - 0x04] == 0x00 && 
				payLoad[0x27 - 0x0E - 0x04] == 0x00 && 
				payLoad[0x28 - 0x0E - 0x04] == 0x01 && 
				payLoad[0x29 - 0x0E - 0x04] == 0xBF); 
		streamID = 0xBF; // change to NAVPack ID
		subStreamID = payLoad[0x2C - 0x0E - 0x04];
		payLoad = &payLoad[0x2D - 0x0E - 0x04];
	}
	else
	{
		subStreamID = 0;
	}

	return 0;
}

// prepares the packet for a new payload with a new payload size
void PESPacket::SetPayLoadSize(int size)
{
	int i;

	// reattach bitstream
	bstream.AttachBuffer(pBuffer, bufferSize);

	// **** ---> see ISO 13818-1 Table 2-18 (PES-Packet) <--- ***
	// **** ---> see ISO 13818-1 Table 2-19 (Stream ID Assignments) <--- ***

	// jump start code
	int val = bstream.ReadBits(32);
	_ASSERT(val >= 0x000001E0 && val <= 0x000001EF);

	bstream.ReadBits(32);
	int header_length = bstream.ReadBits(8);

	_ASSERT(size >= 0);

	// if size is smaller than maximum payload size
	if(size < maxPayLoadSize)
	{
		// if there is enough space for a padding stream
		if(size < maxPayLoadSize - 6)
		{
			_ASSERT(size + 9 + header_length + 6 < maxPayLoadSize + 9 + header_length);
			_ASSERT(maxPayLoadSize + 9 + header_length <= bufferSize);
			// generate padding stream
			pBuffer[size + 9 + header_length] = 0x00;
			pBuffer[size + 9 + header_length + 1] = 0x00;
			pBuffer[size + 9 + header_length + 2] = 0x01;
			pBuffer[size + 9 + header_length + 3] = 0xBE;
			pBuffer[size + 9 + header_length + 4] = (maxPayLoadSize - (size + 6)) >> 8;
			pBuffer[size + 9 + header_length + 5] = (maxPayLoadSize - (size + 6)) & 0xFF;
			for(i = size + 9 + header_length + 6; i < maxPayLoadSize + 9 + header_length; i++)
			{
				pBuffer[i] = 0xFF;
			}
		}
		// if not then add to header length
		else
		{
			int add_length = maxPayLoadSize - size;
			_ASSERT(header_length < 255 - add_length);
			_ASSERT(9 + header_length + add_length <= bufferSize);
			_ASSERT(9 + header_length <= 9 + header_length + add_length);
			for(i = 9 + header_length; i < 9 + header_length + add_length; i++)
			{
				pBuffer[i] = 0xFF;
			}
			header_length += add_length;
			maxPayLoadSize -= add_length;
		}
		// calculate new packet size
		packetSize = size + 3 + header_length;
		// set new packet size
		pBuffer[4] = packetSize >> 8;
		pBuffer[5] = packetSize & 0xFF;
		// recalculate payload pointer
		payLoad = &pBuffer[9 + header_length];
	}
	else
	{
		_ASSERT(FALSE);
		// write packet size
//		bstream.WriteBits(maxPayLoadSize + 9, 16);
		// jump rest
		bstream.ReadBits(16);
		// write 0 header length
//		bstream.WriteBits(0, 8);
		// calculate new packet size
		packetSize = maxPayLoadSize + 9;
		// recalculate payload pointer
		payLoad = &pBuffer[9];
	}

	// set new payload size
	payLoadSize = size;
}
